# menu.tcl --
#
# This file contains an enhanced menu generation system.
#
# Copyright (c) 1996-1998 Donal K. Fellows
#
# 
# The author hereby grants permission to use, copy, modify, distribute,
# and license this software and its documentation for any purpose, provided
# that existing copyright notices are retained in all copies and that this
# notice is included verbatim in any distributions. No written agreement,
# license, or royalty fee is required for any of the authorized uses.
# Modifications to this software may be copyrighted by their authors
# and need not follow the licensing terms described here, provided that
# the new terms are clearly indicated on the first page of each file where
# they apply.
# 
# IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
# DERIVATIVES THEREOF, EVEN IF THE AUTHOR HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# THE AUTHOR AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
# IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHOR AND DISTRIBUTORS HAVE
# NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.
# 
# GOVERNMENT USE: If you are acquiring this software on behalf of the
# U.S. government, the Government shall have only "Restricted Rights"
# in the software and related documentation as defined in the Federal 
# Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
# are acquiring the software on behalf of the Department of Defense, the
# software shall be classified as "Commercial Computer Software" and the
# Government shall have only "Restricted Rights" as defined in Clause
# 252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
# author grants the U.S. Government and others acting in its behalf
# permission to use and distribute the software in accordance with the
# terms specified in this license. 
# 

namespace eval dkfmenu {
    # Mapping from binding sequence specifications to displayed
    # symbol for assorted punctuation characters - I've probably
    # missed loads...

    array set bindname {
        exclam          !
        at              @
        numbersign      #
        dollar          $
        percent         %
        asciicircum     ^
        ampersand       &
        asterisk        *
        parenleft       (
        parenright      )
        minus           -
        equal           =
        backslash       \\
        grave           `
        underscore      _
        plus            +
        bar             |
        asciitilde      ~
        semicolon       ;
        comma           ,
        period          .
        slash           /
        colon           :
        less            <
        greater         >
        question        ?
        braceleft       \{
        braceright      \}
        quotedbl        \"
        bracketleft     \[
        bracketright    \]
        apostrophe      '
    }

    # Convert a binding sequence into a format that is suitable for
    # an accelerator string.  Note that this procedure assumes a
    # well-formed binding sequence.  Since the sequence is also
    # passed to [bind] this isn't too onerous.

    proc readableAccel {sequence} {
        variable bindname

        set accellist {{   }}; # Put some spaces at the beginning
        set ab $sequence
        while {[string length $ab]} {
            # Splitting binding sequences is a little tricky...

            if {[string index $ab 0] == " "} {
                # Get rid of leading spaces, as they are always
                # ignored by [bind]...

                set ab [string trimleft $ab " "]

            } elseif {
                [regexp {^<<([^<>]+)>>(.*)} $ab ? virtual ab]
            } {
                # We have a virtual event - try to remap to real
                # events.  If we manage to map them to an appropriate
                # sequence (i.e. one that includes keypresses) then
                # we will prepend that to the front of the remaining
                # binding sequence and go round the binding process
                # again.  The net effect of this is to make using
                # <<Cut>> as an accelerator come out correctly.

                set substflag 0
                foreach e [event info <<$virtual>>] {
                    if {
                        [string first <   $e] == -1 ||
                        [string first Key $e] > 0
                    } then {
                        set ab "$e $ab"
                        set substflag 1
                        break
                    }
                }
                if {!$substflag} {
                    # No valid substitution (*sigh*) so we do the
                    # best we can so at least the user can see
                    # that something is happening...
                    lappend accellist <<$virtual>>
                }

            } elseif {[regexp {^<([^<>]+)>(.*)} $ab ? bound ab]} {

                # We have a complex binding - make it readable by
                # changing everything into "+"-separated lists of
                # modifiers and the key at the end (letters being
                # capitalised as that looks better)

                set blst {}
                foreach bnd [split $bound -] {
                    switch -regexp $bnd {
                        ^[A-Z]$ {
                            # Special cases - single upper-case
                            # letter
                            lappend blst Shft $bnd
                        }

                        Control {
                            # Special cases - control modifier
                            lappend blst Ctrl
                        }
                        Shift {
                            # Special cases - shift modifier
                            lappend blst Shft
                        }

                        Key[Pp]ress - Key {
                            # Special cases - key type
                            ##### Ignore #####
                        }

                        .+ {
                            # General case - non-empty sequence
                            # element
                            if {[info exist bindname($bnd)]} {
                                lappend blst $bindname($bnd)
                                continue
                            }
                            # Don't know what to do, so capitalise
                            # the first letter, and append.  This is
                            # right most of the time... :^)
                            set c1 [string toupper \
                                    [string index $bnd 0]]
                            set cs [string tolower \
                                    [string range $bnd 1 end]]
                            lappend blst $c1$cs
                        }
                        default {
                            # Special cases - empty sequence element

                            # I can't remember if - can be used as a
                            # non-separator in a binding sequence.
                            # If it can, this catches that case.
                            lappend blst -
                        }
                    }
                }
                lappend accellist [join $blst +]
            } else {
                # Is a simple binding - do the first character only
                # and then go round the loop again to get the next
                # one.  Hey, it works, and I use simple bindings so
                # rarely it doesn't bare thinking about...  :^)

                set chr [string index $ab 0]
                set ab  [string range $ab 1 end]
                if [regexp {[A-Z]} $chr] {
                    lappend accellist "Shft+$chr"
                } else {
                    lappend accellist [string toupper $chr]
                }
            }
        }
        return [string trimright [join $accellist " "]]
    }


    # Takes a label and returns the label without any & symbols in it
    # and the location of the underline (which the first & in the
    # string was in front of.)  & symbols can be placed in the output
    # by quoting them (putting two && signs in a row.)

    proc getLabelAndUnderline {label} {
        # Using a newline in a menu label is a bad idea, so we can
        # get away with this next line to turn quoted & signs into
        # something harmless
        regsub -all {&&} $label "\n" tempLabel

        # Where is the & sign?
        set pos [string first & $tempLabel]

        # This would be easier with [string replace]
        set firstBit  [string range $tempLabel 0 [expr {$pos-1}]]
        set secondBit [string range $tempLabel [expr {$pos+1}] end]

        # Get rid of any extra & signs (there won't be any in firstBit)
        regsub -all {&} $secondBit {} secondBit

        # Restore all the quoted & signs (in unquoted form) and
	# concatenate the bits
        regsub -all "\n" $firstBit$secondBit {\&} newLabel

        # Trap stupidity
        if {$pos>=[string length $newLabel]} {set pos -1}

        # Return the prepared label and the location of the underline
        list $newLabel $pos
    }


    # Procedures to handle context-sensitive help for menus.  Note
    # that the lookup has to be based on both the name of the master
    # menu (known at creation time) and the invoking menu (known at
    # run-time) to ensure that this works correctly even with cloned
    # menus.  Clones are used to implement system menu bars...
    proc handleSelect {varName master this} {
        variable helpMessages
        upvar #0 $varName var

        set key $master###[$this index active]
        if {[info exist helpMessages($key)]} {
            set var $helpMessages($key)
        } else {
            set var {}
        }
    }
    proc addHelpToMenuEntry {menu index message} {
        variable helpMessages
        set helpMessages($menu###$index) $message
    }
    proc removeMenuEntryHelp {menu} {
        variable helpMessages
        foreach key [array names helpMessages $menu###*] {
            unset helpMessages($key)
        }
    }


    # Builds a menu and makes the binding to trigger the help for
    # menu traversal.
    proc makeMenu {menu helpVarName params} {
        eval menu $menu $params
        bind $menu <<MenuSelect>> [namespace code [list \
                handleSelect $helpVarName $menu %W]]
        bind $menu <Destroy> [namespace code [list \
                removeMenuEntryHelp $menu]]
        return $menu
    }

    # Helper procs for adding bits to a menu.  These arrange for all
    # the information conversion and extraction to be done on
    # individual parameters that have been pre-removed (in addition
    # to actually adding the appropriate entry to the menu)

    # Adds an entry (non-cascade) to a menu.  This can adequately
    # handle all the different kinds of entry there are on a menu
    # except for cascades (which need to build their sub-menus)
    proc addEntry {top menu type label accel help params} {
        # What a lot of arguments!
        foreach {lbl ulpos} [getLabelAndUnderline $label] {break}
        eval $menu add $type $params
        set which [$menu index end]
	catch {
	    $menu entryconfigure $which -label $lbl -underline $ulpos
	}
        addHelpToMenuEntry $menu $which $help
        # Only do accelerators at all if there is a toplevel to bind
        # them to...
        if {[string length $top] && [string length $accel]} {
            # Unfortunately, [winfo toplevel $menu] == $menu
            bind $top $accel [list $menu invoke $which]
            $menu entryconfigure $which \
                    -accelerator [readableAccel $accel]
        }
        return $which
    }

    # Adds a cascade entry (and menu) to a menu.
    proc addCascade {
        top menu label helpVarName help subinfo params name
    } {
        # What a lot of arguments!
        foreach {lbl ulpos} [getLabelAndUnderline $label] {break}
        eval $menu add cascade $params
        set which [$menu index end]
        addHelpToMenuEntry $menu $which $help
        if {[string length $name]} {
            set w $menu.$name
        } else {
            set w $menu.m_$which
        }
        set submenu [buildAMenu $top $w $helpVarName $subinfo]
        $menu entryconfigure $which -label $lbl -underline $ulpos \
                -menu $submenu
        return $which
    }


    # Given an array of key/value pairs, look up the given key in the
    # array and return its value (removing that pair from the array.)
    # The default value to return if the value is not present can
    # also be specified.  No abbreviation handling is done, as this
    # makes things easier for me...

    proc getArg {ary key {default {}}} {
        upvar $ary a
        if {[info exist a($key)]} {
            set val $a($key)
            unset a($key)
        } else {
            set val $default
        }
        return $val
    }


    #

    proc buildAMenu {top menu helpVarName info} {
        array set i $info

        set contents [getArg i -contents]

        makeMenu $menu $helpVarName [array get i]

        foreach {type details} $contents {
            set ei(foo) {}
            array set ei $details
            set help  [getArg ei -help]
            set label [getArg ei -label]
            unset ei(foo)
            if {$type == "cascade" && [info exist ei(-submenu)]} {
                set subinfo $ei(-submenu)
                unset ei(-submenu)
                # Name setting allows the definition additions to
                # special system menus (like Apple, Help, or System)
                # Without this option, those menus would remain out
                # of reach with this code (I usually use numeric-
                # based submenu names, as that forces uniqueness.)
                set name [getArg ei -name]
		if {[string index [lindex $subinfo 0] 0] != "-"} {
		    # Add contents if that is all there is!
		    set subinfo [list -contents $subinfo]
		}
                addCascade $top $menu $label $helpVarName $help \
                        $subinfo [array get ei] $name
            } else {
                set accel [getArg ei -accelerator]
                addEntry $top $menu $type $label $accel $help \
                        [array get ei]
            }
            unset ei
        }

        return $menu
    }

    proc createMenubar {toplevel helpvar args} {
        global tcl_platform
        if {$toplevel == "."} {set w {}} {set w $toplevel}
        set menu [buildAMenu $toplevel $w.menu___bar $helpvar $args]
        if {$tcl_platform(platform) == "windows"} {
            # Why must it be done like this?  How should I know!?
            after idle $toplevel configure -menu $menu
        } else {
            $toplevel configure -menu $menu
        }
        return $menu
    }

    proc createUnacceleratedMenu {menuWindow helpvar args} {
        buildAMenu {} $menuWindow $helpvar $args
    }

    namespace export createMenubar createUnacceleratedMenu
}
namespace import ::dkfmenu::*

